﻿/*  
 Copyright (C) 2010 - 2011  Quantum Nomad LLC (http://www.qnomad.com/)

 This program is licensed under version 3 of the GNU General Public License as published
 by the Free Software Foundation.  You should have received a copy of the license along
 with the source code.  If not, an online copy of the license can be found at
 http://www.qnomad.com/wpf/GPL.aspx.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Input;

namespace Qnomad.Outlining
{
	/// <exclude/>
	public enum RelativeInsert { None, Above, Below, Within };

	/// <summary>
	/// Encapsulates an operation to place an item within a list and, if already present in the list,
	/// remove it from its old location. The new position can be described in one of two ways:
	///   (a) as a "nudge" relative to the original position of the item being moved (frontward, rearward), or
	///   (b) as an "insert" relative to another item that acts as a target (sibling after, sibling before).
	/// </summary>
	public class FlatPlacementArgs<TContainer, TItem> : EventArgs
	  where TContainer : IContainer<TContainer, TItem>
	  where TItem : IContained<TContainer, TItem>
	{

		#region Data

		/// <summary>
		/// The present position of the item-to-move
		/// </summary>
		public RelativePosition<TContainer, TItem> oldPosition;

		/// <summary>
		/// The new position of the item-to-move
		/// </summary>
		public RelativePosition<TContainer, TItem> newPosition =
			new RelativePosition<TContainer, TItem>() { childIndex = -1 };

		/// <summary>Is the target list sorted?</summary>
		public readonly bool isSorted;

		public bool IsNewPositionValid { get; private set; }

		/// <summary>Are the arguments ready to be applied?</summary>
		public bool IsValid
		{
			get { return IsNewPositionValid && _itemToMove != null; }
		}

		#endregion Data

		#region Constructors

		/// <summary>
		/// Constructs this from raw drag-and-drop input.
		/// </summary>
		/// <param name="relativeInsert">Where, relative to the target location, to drop</param>
		/// <param name="itemToMove">The item being moved</param>
		/// <param name="targetLoc">The target location of the drop</param>
		/// <param name="isSorted">Is the destination list sorted?</param>
		public FlatPlacementArgs( RelativeInsert relativeInsert, TItem itemToMove, TItem targetLoc, bool isSorted ) {
			this.IsNewPositionValid = false; //false until newPosition is completely set

			// Check for invalid arguments.
			if ( relativeInsert == RelativeInsert.None )
				return; //invalid direction
			else if ( itemToMove == null )
				return; //nothing to move
			else if ( targetLoc == null )
				return; //nowhere to move to
			else if ( itemToMove.Equals(targetLoc) )
				return; //no point in dropping an item back onto itself
			else if ( isSorted && relativeInsert != RelativeInsert.Within )
				return; //cannot insert at specific position within a sorted list

			this.isSorted = isSorted;
			
			// Determine new parent and index.
			int newIndex = targetLoc.Parent.SubItems.IndexOf( targetLoc ); //for Above/Below inserts only
			if ( relativeInsert == RelativeInsert.Above )
				newPosition = new RelativePosition<TContainer, TItem>( targetLoc.Parent, newIndex );
			else if ( relativeInsert == RelativeInsert.Below )
				newPosition = new RelativePosition<TContainer, TItem>( targetLoc.Parent, ++newIndex );
			else if ( relativeInsert == RelativeInsert.Within ) {
				if ( !ProcessWithin(itemToMove, targetLoc) )
					return;
			} else
				throw new InvalidOperationException( "Unknown insert direction!" );

			this.ItemToMove = itemToMove;
			IsNewPositionValid = true;
		}

		protected virtual bool ProcessWithin( TItem itemToMove, TItem targetLoc ) { return false; }

		/// <summary>
		/// Constructs this from raw nudge/insert command input.
		/// </summary>
		/// <param name="command">The nudge or insert command being invoked</param>
		/// <param name="target">The target of the command</param>
		/// <param name="isSorted">Is the destination list sorted?</param>
		public FlatPlacementArgs( ICommand command, TItem target, bool isSorted )
			: this( command, new List<TItem> {target}, isSorted ) { }

		/// <summary>
		/// Constructs this from raw nudge/insert command input.
		/// </summary>
		/// <param name="command">The nudge or insert command being invoked</param>
		/// <param name="targets">The targets of the command</param>
		/// <param name="isSorted">Is the destination list sorted?</param>
		public FlatPlacementArgs( ICommand command, IList<TItem> targets, bool isSorted )
		{
			var nudges = OutliningCommands.Nudges;
			var inserts = OutliningCommands.Inserts;
			if ( !nudges.Union(inserts).Contains(command as RoutedUICommand) )
				throw new ArgumentException( "Only Nudge and Insert commands are valid", "command" );

			if ( GetType() == typeof(FlatPlacementArgs<TContainer, TItem>) ) { //if not hierarchical
				nudges.Remove( OutliningCommands.NudgeDemote );
				nudges.Remove( OutliningCommands.NudgePromote );
				inserts.Remove( OutliningCommands.InsertChild );
				inserts.Remove( OutliningCommands.InsertParent );
			}

			this.isSorted = isSorted;
			IsNewPositionValid = nudges.Contains( command as RoutedUICommand ) ? SetNudgePosition( command, targets[0] ) :
				(inserts.Contains( command as RoutedUICommand ) ? SetInsertPosition( command, targets ) : false);
		}

		#endregion Constructors

		#region 'ItemToMove' Property

		private TItem _itemToMove;

		public TItem ItemToMove
		{
			get { return _itemToMove; }
			set {
				if ( object.Equals(_itemToMove, value) )
					return;
				else if ( _itemToMove != null )
					throw new InvalidOperationException( "Cannot overwrite a previously set ItemToMove" );

				itemToMove = value;
				SetSortedIndex();

				// Adjust new index, if necessary.
				if ( this.oldPosition.parent != null && this.newPosition.parent != null &&
						this.oldPosition.parent.Equals( this.newPosition.parent ) &&
						this.oldPosition.childIndex < this.newPosition.childIndex )
					this.newPosition.childIndex--;
			}
		}

		private TItem itemToMove
		{
			set
			{
				_itemToMove = value;

				// Determine former parent and index, if already positioned.
				if ( _itemToMove.Parent != null ) {
					IList<TItem> oldSubItems = _itemToMove.Parent.SubItems;
					this.oldPosition = new RelativePosition<TContainer, TItem>( _itemToMove.Parent, oldSubItems.IndexOf( _itemToMove ) );
				}
			}
		}

		protected void SetSortedIndex()
		{
			if ( !isSorted || _itemToMove == null )
				return;

			var subItems = this.newPosition.parent.SubItems;
			var targetLoc = subItems.Cast<IComparable<TItem>>().FirstOrDefault(
				c => c.CompareTo( _itemToMove ) >= 0 && !c.Equals( _itemToMove )
			);
			this.newPosition.childIndex = targetLoc != null ? subItems.IndexOf( (TItem)targetLoc ) : subItems.Count;
		}

		#endregion ItemToMove

		/// <summary>
		/// Determines an item's new position based on a nudge command and the item's old position.
		/// </summary>
		/// <param name="command">The nudge command</param>
		/// <param name="target">The item to reposition</param>
		/// <returns>False if arguments are invalid; true otherwise</returns>
		protected virtual bool SetNudgePosition( ICommand command, TItem target )
		{
			if ( target == null || target.Parent == null || //cannot determine new position w/o knowing old position
					(isSorted && //cannot move to a specific index within a sorted list
						(command == OutliningCommands.NudgeFrontward || command == OutliningCommands.NudgeRearward)) )
				return false;

			this.itemToMove = target;

			if ( command == OutliningCommands.NudgeFrontward ) {
				if ( oldPosition.childIndex == oldPosition.parent.SubItems.Count - 1 ) return false; //Already at the end!
				newPosition = new RelativePosition<TContainer, TItem>( oldPosition.parent, oldPosition.childIndex + 1 );
			} else if ( command == OutliningCommands.NudgeRearward ) {
				if ( oldPosition.childIndex == 0 ) return false; //Already at the start!
				newPosition = new RelativePosition<TContainer, TItem>( oldPosition.parent, oldPosition.childIndex - 1 );
			}
			return true;
		}

		/// <summary>
		/// Determines an item's new position based on an insert command and its targets.
		/// </summary>
		/// <param name="command">The insert command</param>
		/// <param name="targets">The targets</param>
		/// <returns>False if arguments are invalid; true otherwise</returns>
		protected virtual bool SetInsertPosition( ICommand command, IList<TItem> targets )
		{
			if ( targets == null || targets.Count == 0 || targets[0] == null || //unknown position
					((command == OutliningCommands.InsertSiblingBefore || command == OutliningCommands.InsertSiblingAfter) &&
						(targets[0].Parent == null || //cannot become a sibling of a root
							isSorted) //cannot insert at specific position within a sorted list
					)
				)
				return false;

			if ( command == OutliningCommands.InsertSiblingBefore || command == OutliningCommands.InsertSiblingAfter ) {
				this.newPosition.parent = targets[0].Parent; //decide parent of new item

				//Decide index at which to insert the new item:
				TContainer newParent = this.newPosition.parent;
				IList<TItem> subItems = newParent != null ? newParent.SubItems : null;
				this.newPosition.childIndex =
					command == OutliningCommands.InsertSiblingBefore ? MinimumIndex( subItems, targets ) :
					command == OutliningCommands.InsertSiblingAfter ? MaximumIndex( subItems, targets ) + 1 :
					-1;
			}
			return true;
		}

		#region Helpers

		protected static int MinimumIndex( IList<TItem> parentList, IList<TItem> subItems )
		{
			int result = parentList.Count;
			foreach ( TItem item in subItems ) {
				int index = parentList.IndexOf( item );
				if ( index < result ) result = index;
			}
			return result;
		}

		protected static int MaximumIndex( IList<TItem> parentList, IList<TItem> subItems )
		{
			int result = 0;
			foreach ( TItem item in subItems ) {
				int index = parentList.IndexOf( item );
				if ( index > result ) result = index;
			}
			return result;
		}

		public override string ToString()
		{
			return GetType().Name + ": parent={" + newPosition.parent + "}";
		}

		#endregion Helpers

		#region Execute Method

		public virtual void Apply() {
			if ( !IsValid )
				throw new InvalidOperationException( "Invalid arguments!" );

			if ( oldPosition.parent != null )
				oldPosition.parent.SubItems.RemoveAt( oldPosition.childIndex ); //remove the item

			if ( newPosition.childIndex >= 0 )
				newPosition.parent.SubItems.Insert( newPosition.childIndex, _itemToMove ); //re-add the removed item
			else
				newPosition.parent.SubItems.Add( _itemToMove ); //re-add the removed item

			IsNewPositionValid = false;
		}

		#endregion Execute Method
	}
}
