﻿/*  
 Copyright (C) 2010 - 2011  Quantum Nomad LLC (http://www.qnomad.com/)

 This program is licensed under version 3 of the GNU General Public License as published
 by the Free Software Foundation.  You should have received a copy of the license along
 with the source code.  If not, an online copy of the license can be found at
 http://www.qnomad.com/wpf/GPL.aspx.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;

namespace Qnomad.Outlining {

	/// <seealso href="http://www.codeproject.com/KB/recipes/nonintrusive-trees-graphs.aspx"/>
	public static class ITreeNodeExtensions
	{
		#region Tree Operations

		public static IEnumerable<T> Ancestors<T>( this ITreeNode<T> startNode )
			where T : ITreeNode<T>
		{
			return TreeOperations.Ancestors( startNode, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		public static IEnumerable<T> AncestorsAndSelf<T>( this ITreeNode<T> startNode )
			where T : ITreeNode<T>
		{
			return TreeOperations.AncestorsAndSelf( startNode, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		/// <summary>
		/// Returns an iterator that performs a breadth-first traversal of nodes.
		/// </summary>
		public static IEnumerable<T> Descendants<T>( this ITreeNode<T> root )
		  where T : ITreeNode<T>
		{
			return TreeOperations.Descendants( root, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		/// <summary>
		/// Returns an iterator that performs a breadth-first traversal of nodes.
		/// </summary>
		public static IEnumerable<T> DescendantsAndSelf<T>( this ITreeNode<T> root )
		  where T : ITreeNode<T>
		{
			return TreeOperations.DescendantsAndSelf( root, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		public static IEnumerable<T> PrecedingSiblings<T>( this ITreeNode<T> context )
		  where T : ITreeNode<T>
		{
			return TreeOperations.PrecedingSiblings( context, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		public static IEnumerable<T> FollowingSiblings<T>( this ITreeNode<T> context )
		  where T : ITreeNode<T>
		{
			return TreeOperations.FollowingSiblings( context, ITreeNodeAdapterImpl<T>.Instance ).Cast<T>();
		}

		#endregion Tree Operations

		#region 'GetParallelNode' Method

		/// <summary>
		/// Gets the path leading from the tree root to the supplied node.
		/// </summary>
		/// <param name="target">A tree node</param>
		/// <returns>A stack whose topmost element is the tree root, and whose bottommost is the specified node</returns>
		public static Stack<T> GetPathFromRoot<T>( this T target )
		  where T : ITreeNode<T>
		{
			if ( target == null )
				throw new ArgumentNullException( "target" );

			Stack<T> ancestors = new Stack<T>();
			for ( T i = target; i != null; i = i.Parent )
				ancestors.Push( i );
			return ancestors;
		}

		/// <summary>
		/// Gets the corresponding node in a parallel tree.
		/// </summary>
		/// <typeparam name="T">Source tree's type</typeparam>
		/// <typeparam name="U">Target tree's type</typeparam>
		/// <param name="nodeInT">Destination in the source tree</param>
		/// <param name="rootInU">Root of the target tree</param>
		/// <param name="isMapped">Indicates if two nodes correlate or not</param>
		/// <returns>The node in U that correlates to the node in T</returns>
		public static U GetParallelNode<T, U>( this T nodeInT, U rootInU, Func<U, T, bool> isMapped )
		  where T : ITreeNode<T>
		  where U : class, ITreeNode<U>
		{
			Stack<T> pathInT = nodeInT.GetPathFromRoot();
			if ( rootInU == null || !isMapped( rootInU, pathInT.Pop() ) )
				return null; //roots do not correlate

			U nodeInU = rootInU;
			while ( pathInT.Count > 0 ) {
				T t = pathInT.Pop();
				nodeInU = nodeInU.SubItems.Single( u => isMapped( u, t ) );
			}
			return nodeInU;
		}

		#endregion GetParallelNode

		#region 'GetTreeViewItem' Method

		/// <summary>
		/// Navigates from treeView to the TreeViewItem which corresponds to "this".
		/// </summary>
		/// <param name="treeView">The tree root in the UI</param>
		/// <param name="callback">A delegate that will be passed the TreeViewItem
		///   corresponding to "this" once that tvi becomes available</param>
		///   <returns>Indicates whether or not the navigation could begin</returns>
		public static bool GetTreeViewItem<T>( this T dataContext, TreeView treeView, Action<TreeViewItem> callback )
		  where T : ITreeNode<T>
		{
			string missingParam = treeView == null ? "treeView" : (callback == null ? "callback" : null);
			if ( missingParam != null ) //check for required parameters
				throw new ArgumentNullException( missingParam );
			else if ( dataContext == null ) //check for optional parameter
				return false;

			new StatusChangedListener<T>( treeView, dataContext, callback );
			return true;
		}

		public static bool FocusTreeViewItem<T>( this T dataContext, TreeView treeView )
		  where T : ITreeNode<T>
		{
			return GetTreeViewItem<T>( dataContext, treeView, tvi => {
				if ( tvi != null ) //if the corresponding tvi still exists...
					tvi.Focus(); //then focus it
			} );
		}

		class StatusChangedListener<T>
		  where T : ITreeNode<T>
		{
			Stack<T> ancestors;
			ItemsControl ic;
			T target;
			Action<TreeViewItem> callback;

			public StatusChangedListener( TreeView treeView, T target, Action<TreeViewItem> callback )
			{
				this.ic = treeView; this.target = target; this.callback = callback;

				ancestors = new Stack<T>();
				for ( T i = target; i != null; i = i.Parent ) {
					ancestors.Push( i );
					if ( treeView.Items.Contains( i ) ) //is at top level?
						break;
				}
				Continue();
			}

			void Continue()
			{
				while ( ancestors.Count > 0 ) {
					if ( ic.ItemContainerGenerator.Status == GeneratorStatus.ContainersGenerated ) {
						T pm = ancestors.Pop();
						ic = (ItemsControl)ic.ItemContainerGenerator.ContainerFromItem( pm );
					} else {
						ic.ItemContainerGenerator.StatusChanged += StatusChanged;
						return;
					}
				}
				callback( (TreeViewItem)ic );
			}

			void StatusChanged( object sender, EventArgs e )
			{
				if ( ic.ItemContainerGenerator.Status == GeneratorStatus.ContainersGenerated ) {
					ic.ItemContainerGenerator.StatusChanged -= StatusChanged;
					Continue();
				}
			}
		}

		#endregion

		public static int PositionInList<T>( this ITreeNode<T> item )
		  where T : ITreeNode<T>
		{
			return item.Parent == null ? 0 :
				item.Parent.SubItems.IndexOf( (T)item );
		}

		public static T Root<T>( this ITreeNode<T> item )
		  where T : ITreeNode<T>
		{
			return item.AncestorsAndSelf().Single( pm => pm.Parent == null );
		}
	}
}
