﻿using System;
using System.Diagnostics;
using System.Windows.Input;

namespace Qnomad.PresentationModel
{
	/// <summary>
	/// A command whose sole purpose is to relay its functionality to other objects by
	/// invoking delegates. The default return value for the CanExecute method is 'true'.
	/// </summary>
	/// <seealso href="http://msdn.microsoft.com/en-us/magazine/dd419663.aspx"/>
	public class DelegateCommand : ICommand
	{
		#region Fields

		readonly Action<object> _execute;
		readonly Predicate<object> _canExecute;

		#endregion // Fields

		#region Constructors

		/// <summary>
		/// Creates a new command that can always execute.
		/// </summary>
		/// <param name="execute">The execution logic.</param>
		public DelegateCommand( Action<object> execute )
			: this( execute, null ) { }

		/// <summary>
		/// Creates a new command.
		/// </summary>
		/// <param name="execute">The execution logic.</param>
		/// <param name="canExecute">The execution status logic.</param>
		public DelegateCommand( Action<object> execute, Predicate<object> canExecute )
		{
			if ( execute == null )
				throw new ArgumentNullException( "execute" );

			_execute = execute;
			_canExecute = canExecute;
		}

		#endregion // Constructors

		#region ICommand Members

		[DebuggerStepThrough]
		public bool CanExecute( object parameter )
		{
			return _canExecute == null ? true : _canExecute( parameter );
		}

		public event EventHandler CanExecuteChanged
		{
			add { CommandManager.RequerySuggested += value; }
			remove { CommandManager.RequerySuggested -= value; }
		}

		public void Execute( object parameter )
		{
			_execute( parameter );
		}

		#endregion // ICommand Members
	}
}
