/*  
 Copyright (C) 2010 - 2011  Quantum Nomad LLC (http://www.qnomad.com/)

 This program is licensed under version 3 of the GNU General Public License as published
 by the Free Software Foundation.  You should have received a copy of the license along
 with the source code.  If not, an online copy of the license can be found at
 http://www.qnomad.com/wpf/GPL.aspx.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Linq;

using Qnomad.BidirectionalAssociations;
using Qnomad.Outlining;

namespace Qnomad.PresentationModel
{
	/// <summary>Presentation-tailored hierarchical representation of a domain-model type</summary>
	/// <typeparam name="DMType">The domain-model type to be represented</typeparam>
	/// <typeparam name="PMType">The presentation-model type (this) doing the representing</typeparam>
	public abstract class HierarchicalPM<DMType, PMType> : ItemPM<DMType>, ITreeNode<PMType>, IComparable<PMType>
	  where DMType : class
	  where PMType : HierarchicalPM<DMType, PMType>
	{
		#region Object Creation

		private bool useSourceIndex, cascadeDeletes;
		protected bool debug;

		/// <summary>
		/// Initializes this presentation-model node with the domain items to be
		/// represented, then subscribes for notification of future changes.
		/// </summary>
		/// <param name="domainItem">The domain item to be represented by this node</param>
		/// <param name="subItemsSource">The domain items to be represented as this node's sub-items.
		///   Null if root wrapper will be responsible for all add operations.</param>
		/// <param name="useSourceIndex">Specifies whether presentation-model subitems' positions here,
		///   are the same as in their corresponding subItemsSource</param>
		/// <param name="cascadeDeletes">Specifies whether subitems also get deleted once their parent is deleted.
		///   If false, then a deleted item's subitems are re-added at their parent's former location.</param>
		protected HierarchicalPM( DMType domainItem, ICollection<DMType> subItemsSource,
		  bool useSourceIndex, bool cascadeDeletes, bool debug ) : base( domainItem )
		{
			this.SubItemsSource = subItemsSource;
			this.useSourceIndex = useSourceIndex; this.cascadeDeletes = cascadeDeletes; this.debug = debug;
		}

		/// <summary>
		/// Creates a new presentation-model item.
		/// </summary>
		/// <param name="domainItem">The domain item to be represented by the new instance</param>
		/// <returns>The newly-created instance</returns>
		protected abstract PMType CreateInstance( DMType domainItem );

		#endregion

		#region SubItems/Parent Properties

		private ICollection<DMType> subItemsSource;
		private ObservableCollection<PMType> _subItems;
		private ReadOnlyObservableCollection<PMType> subItemsAsReadOnly;
		private PMType parent;

		protected ICollection<DMType> SubItemsSource
		{
			set
			{
				UnsubscribeFromSource();
				if ( HasSubItems ) subItems.Clear();
				subItemsSource = value;
				SubscribeToSource(); //subscribe to source's change events
			}
		}

		protected ObservableCollection<PMType> subItems
		{
			get
			{
				if ( _subItems == null ) {
					_subItems = new ObservableCollection<PMType>();
					( (INotifyCollectionChanged)_subItems ).CollectionChanged +=
						new OneToManyAssocSync( this, "Parent" ).UpdateManySide;
				}
				return _subItems;
			}
		}

		public bool HasSubItems
		{
			get { return _subItems != null && subItems.Count > 0; }
		}

		/// <summary>The sub-items of this presentation-model node, as a read-only collection</summary>
		/// <remarks>
		/// It is the responsibility of this object to ensure that this sub-items collection reflects 
		/// the contents of the corresponding domain-model collection.  It does this by listening for 
		/// changes to the source collection and updating itself accordingly.  It does not, however, 
		/// propagate changes in the other direction. This sub-items collection is therefore available 
		/// to the outside only in read-only form.
		/// </remarks>
		public ReadOnlyObservableCollection<PMType> SubItems
		{
			get
			{
				if ( subItemsAsReadOnly == null )
					subItemsAsReadOnly = new ReadOnlyObservableCollection<PMType>( subItems );
				return subItemsAsReadOnly;
			}
		}

		public PMType Parent
		{
			get { return parent; }
			private set
			{
				if ( value == parent ) return; //ignore if not a real change

				PMType oldParent = parent;
				IList<PMType> selectedInSubtree = null;

				if ( value == null ) { //if detaching a subtree...
					selectedInSubtree = this.DescendantsAndSelf().Where( pm => pm.IsSelected ).ToList();
					if ( selectedInSubtree.Count > 0 ) { //and if it contains selected items...
						foreach ( PMType pm in selectedInSubtree )
							pm.IsSelected = false; //then deselect to avoid 'multiple selections' error for deletes
					}
				} else if ( selected != null && selected.Count > 0 ) { //if (re)attaching subtree that has root data...
					var root = value.Root();
					if ( root.selected != null ) { //and if new root also has data...
						foreach ( PMType pm in root.selected.ToList() )
							pm.IsSelected = false; //then clear data on new root
					}
				}

				parent = value;
				OneToManyAssocSync.UpdateOneSide<PMType>( (PMType)this, oldParent, parent, "subItems" );

				if ( parent == null ) { //if detached a subtree...
					foreach ( PMType pm in selectedInSubtree ) //and if it contained selected items...
						pm.IsSelected = true; //then restore original IsSelected value in case subtree is re-added
				} else if ( selected != null ) { //if (re)attached subtree that has root data...
					foreach ( PMType pm in selected )
						pm.OnIsSelectedChanged( true ); //transfer subtree's root data to new root
					selected = null; //and clear data on subtree
				}
			}
		}

		#endregion

		#region Positioning Sub-Items

		/// <summary>
		/// Gets the desired position for a presentation-model item.
		/// </summary>
		/// <param name="subtree">The subtree within which to place this item</param>
		protected abstract HierarchicalPlacementArgs<PMType> DesiredPosition( PMType subtree );

		/// <summary>
		/// Gets the final position at which to add a presentation-model descendant.
		/// </summary>
		/// <param name="subtree">The subtree within which to place this item</param>
		/// <param name="sourceIndex">The index at which to add the child item, if applicable</param>
		protected HierarchicalPlacementArgs<PMType> PreparePosition( PMType subtree, int sourceIndex = -1 )
		{
			var result = DesiredPosition( subtree );
			result.ItemToMove = (PMType)this;
			if ( useSourceIndex && sourceIndex >= 0 ) result.newPosition.childIndex = sourceIndex;
			if ( debug ) Debug.WriteLine( "Adding {" + DomainItem + "}, Positioning {" + result + "}" );
			return result;
		}

		public void UpdatePosition()
		{
			PMType ancestor = this.Ancestors().First( pm => pm.subItemsSource != null );
			var desiredPos = PreparePosition( ancestor );

			if ( desiredPos.itemsToSubsume == null && subItems.Count > 0 ) {
				foreach ( PMType pm in subItems.ToList() )
					pm.PreparePosition( Parent ).Apply();
				desiredPos = PreparePosition( ancestor ); //new index may have changed
			}
			if ( Parent != desiredPos.newPosition.parent ||
					this.PositionInList() != desiredPos.newPosition.childIndex )
				desiredPos.Apply();
			else if ( desiredPos.itemsToSubsume != null &&
					!subItems.EqualsIgnoreOrder(desiredPos.itemsToSubsume) ) {
				foreach ( PMType pm in subItems.ToList() ) {
					if ( !desiredPos.itemsToSubsume.Contains(pm) )
						pm.PreparePosition( Parent ).Apply();
				}
				foreach ( PMType pm in desiredPos.itemsToSubsume )
					pm.PreparePosition( (PMType)this ).Apply();
			}
		}

		public virtual int CompareTo( PMType other )
		{
			var dm = DomainItem as IComparable<DMType>;
			if ( dm != null )
				return dm.CompareTo( other.DomainItem );
			else
				throw new NotSupportedException( "The domain object must implement IComparable<DMType>" );
		}

		#endregion

		#region 'SubItems' Change Notification

		/// <summary>
		/// The last sub-item removed; remains available until another sub-item is added.
		/// Useful when the item is being moved from one location (where it's just been
		/// deleted) in the hierarchy, to another (where it will next be re-added).
		/// </summary>
		private static PMType lastRemoved;

		/// <summary>
		/// Responds to the represented domain item's sub-items collection changing.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void SubItemsSourceChanged( object sender, NotifyCollectionChangedEventArgs args ) {
			if ( args.Action == NotifyCollectionChangedAction.Add ) {
				//foreach ( DMType addingToCollection in e.NewItems )
				for ( int i = 0; i < args.NewItems.Count; i++ ) {
					DMType addingToCollection = (DMType)args.NewItems[i];
					if ( lastRemoved != null && lastRemoved.DomainItem == addingToCollection ) {
						lastRemoved.PreparePosition( (PMType)this, args.NewStartingIndex + i ).Apply();
						lastRemoved.SubscribeToSource();
					} else
						CreateInstance( addingToCollection ).PreparePosition( (PMType)this, args.NewStartingIndex + i ).Apply();
					lastRemoved = null;
				}
			} else if ( args.Action == NotifyCollectionChangedAction.Remove ) {
				foreach ( DMType removingFromCollection in args.OldItems ) {
					PMType pm = FindDescendant( removingFromCollection );
					int lastRemovedIndex = pm.Parent.SubItems.IndexOf( pm );
					pm.Parent = null;
					pm.UnsubscribeFromSource();
					lastRemoved = pm;

					//When deletes don't cascade, re-add a deleted item's subitems:
					if ( !cascadeDeletes ) {
						IList<PMType> toReAdd = new List<PMType>( pm.SubItems );
						for ( int i = 0; i < toReAdd.Count; i++ )
							toReAdd[i].PreparePosition( (PMType)this, lastRemovedIndex + i ).Apply();
					}
				}
			}
		}

		/// <summary>
		/// Subscribe to the sub-items source's change events.
		/// </summary>
		private void SubscribeToSource() {
			if ( subItemsSource != null )
				( (INotifyCollectionChanged)subItemsSource ).CollectionChanged += SubItemsSourceChanged;
		}

		/// <summary>
		/// Unsubscribe from the sub-items source's change events.
		/// </summary>
		private void UnsubscribeFromSource() {
			if ( subItemsSource != null )
				( (INotifyCollectionChanged)subItemsSource ).CollectionChanged -= SubItemsSourceChanged;
		}

		#endregion

		#region 'IsExpanded' Property

		private bool isExpanded;

		/// <summary>
		/// Is the TreeViewItem associated with this object expanded?
		/// </summary>
		/// <seealso href="http://www.codeproject.com/KB/WPF/TreeViewWithViewModel.aspx"/>
		public bool IsExpanded {
			get { return isExpanded; }
			set {
				SetProperty( ref isExpanded, value, "IsExpanded" );

				// Expand all the way up to the root:
				if ( isExpanded && Parent != null )
					Parent.IsExpanded = true;
			}
		}

		#endregion // IsExpanded

		#region 'SelectedItem' Property

		IList<PMType> selected;
		PMType prevSelected; //for which change events were raised
		event EventHandler<SelectionChangedEventArgs<DMType, PMType>> selectionChanged;

		public PMType SelectedItem
		{
			get
			{
				VerifyIsRoot();
				if ( selected == null ) return null;

				switch ( selected.Count ) {
					case 0: return null;
					case 1: return selected[0];
					default:
						var s = string.Join( " ,", selected.Select( pm => pm.ToString() ).ToArray() );
						throw new InvalidOperationException( "Multiple items selected: " + s );
				}
			}
		}

		public bool SuppressSelectionChangedEvents { get; set; }

		public event EventHandler<SelectionChangedEventArgs<DMType, PMType>> SelectionChanged
		{
			add { VerifyIsRoot(); selectionChanged += value; }
			remove { VerifyIsRoot(); selectionChanged -= value; }
		}

		protected override void OnIsSelectedChanged( bool newValue )
		{
			this.Root().HandleIsSelectedChanged( (PMType)this, newValue );
		}

		void HandleIsSelectedChanged( PMType changedPM, bool newValue )
		{
			VerifyIsRoot();

			// Update stored selection.
			if ( selected == null ) selected = new List<PMType>();
			if ( newValue )
				selected.Add( changedPM );
			else
				selected.Remove( changedPM );

			// Raise events if selection has changed.
			PMType newSelection = null;

			if ( selected.Count == 1 ) {
				newSelection = selected[0];
				if ( newSelection.Parent != null )
					newSelection.Parent.IsExpanded = true; //ensure item is visible
			}

			if ( selected.Count < 2 && newSelection != prevSelected ) { //if selection really changed...
				if ( SuppressSelectionChangedEvents )
					SuppressSelectionChangedEvents = false;
				else {
					OnPropertyChanged( "SelectedItem" ); //raise PropertyChanged event
					if ( selectionChanged != null ) {
						var args = new SelectionChangedEventArgs<DMType, PMType>( newSelection );
						selectionChanged( this, args ); //raise SelectionChanged event
					}
					prevSelected = newSelection;
				}
			}
		}

		void VerifyIsRoot()
		{
			if ( Parent != null )
				throw new InvalidOperationException( "Operation is only valid on a root" );
		}

		#endregion SelectedItem

		#region Search/Traversal Methods

		/// <summary>
		/// Search within this presentation-model node, and its sub-nodes, using the supplied criteria.
		/// </summary>
		/// <typeparam name="T">Only check nodes of this type</typeparam>
		/// <param name="testNode">Criteria by which to choose which node to return</param>
		/// <returns>First node found that matches the given criteria</returns>
		public PMType FindDescendant<T>( Predicate<T> testNode ) where T : class, DMType {
			/*foreach ( PMType node in this.BreadthFirstNodeEnumerator() ) {
				T domainItemAsT = node.DomainItem as T;
				if ( domainItemAsT != null && testNode(domainItemAsT) ) return node;
			}
			return null;*/
			return ( from PMType pm in this.DescendantsAndSelf()
						let dmAsT = pm.DomainItem as T
						where dmAsT != null && testNode( dmAsT )
						select pm
				   ).FirstOrDefault();
		}

		/// <summary>
		/// Search within this presentation-model node, and its sub-nodes, for the specified domain item.
		/// </summary>
		/// <param name="forDomainItem">
		///   Domain item for which to find a corresponding presentation-model</param>
		/// <returns>The presentation-model that represents the specified domain item</returns>
		public PMType FindDescendant( DMType forDomainItem ) {
			Predicate<DMType> testNode = delegate( DMType nodeToTest ) {
				return nodeToTest == forDomainItem;
			};
			return FindDescendant( testNode );
		}

		/// <summary>
		/// Search within this presentation-model node, and its sub-nodes, using the supplied criteria.
		/// </summary>
		/// <typeparam name="T">Only check nodes of this type</typeparam>
		/// <param name="testNode">Criteria by which to choose which nodes to return</param>
		/// <returns>All nodes found that match the given criteria</returns>
		public IList<PMType> FindDescendants<T>( Predicate<T> testNode ) where T : class, DMType {
			return ( from PMType pm in this.DescendantsAndSelf()
						let dmAsT = pm.DomainItem as T
						where dmAsT != null && testNode( dmAsT )
						select pm
				   ).ToList();
		}

		#endregion

		#region ITreeNode Members

		PMType IContained<PMType, PMType>.Parent {
			get { return Parent; }
		}

		IList<PMType> IContainer<PMType, PMType>.SubItems {
			get { return subItems; }
		}

		#endregion
	}
}
